#!/usr/bin/runhaskell
{-# LANGUAGE FlexibleContexts #-}

import Prelude hiding (fail, uncurry)
import Tct.Configuration
import Tct.Interactive
import Tct.Instances
-- import Tct.Th hiding (config)

import qualified Data.Set as Set

import Tct.Processor.Args
import qualified Tct.Instances as Instance
import qualified Tct.Processors as Processor
import qualified Tct.Processor as P
import qualified Tct.Processor.Standard as S
import qualified Tct.Certificate as Cert
import qualified Tct.Method.DP.DependencyGraph as DG

import Termlib.Utils (snub)

import qualified Termlib.Repl as TR
import Abbrevs

import qualified Raml as Raml


main :: IO ()
main = tct config

config :: Config
config = defaultConfig { strategies = strats
                       , interactiveShowProofs = True
                       , recompile = False } 
  where
    strats = 
      [ Raml.raml  ::: strategy "raml" (optional (maybeArg naturalArg) "inner-timeout" (Just 15)
                                        :+: optional boolArg "greedy" False
                                        :+: optional boolArg "smallbits" True                                        
                                        :+: optional boolArg "bigbits" True
                                        :+: optional naturalArg "semdegree" (Nat 2))
      , mxs        ::: strategy "matrices" (optional naturalArg "mindegree" (Nat 1)
                                            :+: optional naturalArg "maxdegree" (Nat 3))
      , pxs        ::: strategy "polys" (optional naturalArg "mindegree" (Nat 1)
                                         :+: optional naturalArg "maxdegree" (Nat 3))
      , sem        ::: strategy "semantics" (optional naturalArg "mindegree" (Nat 1)
                                             :+: optional naturalArg "maxdegree" (Nat 3))
      , simple     ::: strategy "simple" (optional boolArg "dp" True)                                                 
      , rc2012RCi  ::: strategy "rc2012" 
      ]
          

--------------------------------------------------------------------------------
-- conveniences                
--------------------------------------------------------------------------------

app m = apply m >> return ()
loadDPi s = loadIRC s >> app toDP

loadUncurried fp = 
  do load fp
     apply uncurry
     [p] <- problems
     modifyInitialWith $ const $ f p
    where f p = p { TR.startTerms = TR.BasicTerms ds cs 
                  , TR.strategy = TR.Innermost} 
            where rs = TR.allComponents p
                  ds = TR.definedSymbols rs
                  cs = TR.constructors rs
                

--------------------------------------------------------------------------------
-- combinators
--------------------------------------------------------------------------------


mxs (Nat min :+: Nat max) = fastest $ concat [ [mx i i, mx (i + 1) i] | i <- [min..max] ]
pxs (Nat min :+: Nat max) = fastest [ polys i | i <- [min..max] ]
sem args = mxs args  `orFaster` pxs args

----------------------------------------------------------------------
-- selectors
--------------------------------------------------------------------------------

selAnyLeaf = selAnyOf $ selFromCWDG "strict leaf in CWDG" sel
  where sel cwdg = TR.emptyRuleset { TR.sdp = TR.fromRules [ r | (DG.StrictDP, r) <- leafRules cwdg] }
        leafRules cwdg = DG.allRulesFromNodes cwdg (DG.leafs cwdg)
  

selStrictLeafs = selAllOf $ selFromWDG "all leaf rules from WDG" sel
  where sel wdg = TR.emptyRuleset { TR.sdp = TR.fromRules rs }
          where stricts = [ (n,r) | (n,(DG.StrictDP,r)) <- DG.withNodeLabels' wdg (DG.nodes wdg)]
                rs = [ r | (n, r) <- stricts
                         , all (\ (m, (strictness, _)) -> n == m || strictness == DG.WeakDP) $ DG.withNodeLabels' wdg $ DG.reachablesDfs wdg [n] ]

----------------------------------------------------------------------
-- transformations
--------------------------------------------------------------------------------

splitCongruence = decomposeDG decomposeDGselect

wgOnUsable = wg 2 1 `wgOn` WgOnTrs

removeLeafs 0 = exhaustively $ whenNonTrivial $ removeLeaf (mx 1 0)
removeLeafs i = removeLeaf (spopstar i) 
                <> (removeLeaf (mx i i) 
                    <||> removeLeaf (mx (i + 1) i)
                    <||> when (i == 1) (removeLeaf (mx 3 1))                    
                    <||> when (i == 2) (removeLeaf (polys 2)))
        
isTrivialDP = 
  te removeInapplicable
  >>> try cleanTail 
  >>> try trivial
  >>| empty



----------------------------------------------------------------------
--- 2012 strategies
        
--- innermost
        

directs to = 
  timeout to (te (compse 1)
              >>> te (compse 2)
              >>> te (compse 3)
              >>> te (compse 4)
              >>| empty)                           
  `orBetter` timeout (min 5 to) matchbounds
  `orBetter` timeout to ( bsearch "popstar" popstarSmallPS )
  `orBetter` timeout to popstarPS
  where compse i = withProblem $ \ prob ->
          whenNonTrivial $ 
            decomposeDynamic Add (spopstar i) -- binary search auf grad
            <> (when (i == 2 || i == 3) (decomposeDynamic Add (polys i))
                <||> wg i i
                <||> decomposeDynamic Add (mx i i)
                <||> when (i < 4) (decomposeDynamic Add (mx (i + 1) i)))
          
                               
shiftLeafs = removeLeafs 0 <||> (removeLeafs 1 <> removeLeafs 2)

removeFirstCongruence proc = 
  decomposeDG decomposeDGselect `solveUpperWith` (try simps >>| proc) >>> try simps


isTrivialCWDG cwdg = 
  all null $ [ DG.successors cwdg n | n <- DG.roots cwdg ]

rc2012DPi = 
  toDP >>!! te (withCWDG trans) >>! basics
  where 
    trans cwdg 
      | isTrivialCWDG cwdg = some $ shiftLeafs 
      | otherwise = some $ timeout 15 shiftLeafs <> rc
      where 
        proc = try simps >>> try simpDPRHS >>> te shiftLeafs >>! basics
        rc = removeFirstCongruence proc

rc2012RCi = 
  try irr >>! (directs 59 `orBetter` timeout 59 rc2012DPi)

    
peLeaf proc = proc `withPEOn` selAnyLeaf

                   
selSplitCongruence = selAllOf $ selFromCWDG "below first cut in CWDG" fn
  where 
    fn cwdg = 
      case DG.roots cwdg of 
        (r:_) -> selBelow cwdg r
        _ -> TR.emptyRuleset
    selBelow cwdg r = TR.emptyRuleset { TR.sdp = TR.fromRules [ rl | (DG.StrictDP, rl) <- rs ]
                                      , TR.wdp = TR.fromRules [ rl | (DG.WeakDP, rl) <- rs ] }
      where 
        rs = DG.allRulesFromNodes cwdg $ DG.successors cwdg r
                   
----------------------------------------------------------------------
--- ZK setup
----------------------------------------------------------------------                

simple False = 
  successive [te (rl i) | i <- [1..4]] >>! basics
  where 
    rl 0 = whenNonTrivial $ cmp (mx 1 0)
    rl i = whenNonTrivial $ 
           cmp (mx i i)
           <||> cmp (mx (i+1) i)
           <||> cmp (spopstar i)
           <||> when (i == 2) (cmp (polys 2))
    cmp p = decomposeDynamic Add p
simple True = 
  toDP >>> te shift >>! basics 
  where 
    shift = rl 1 <> rl 2 <> rl 3 <> rl 4
    rl 0 = whenNonTrivial $ removeLeaf (mx 1 0)
    rl i = whenNonTrivial $ removeLeaf (spopstar i)
                            <> (removeLeaf (mx i i) 
                                <||> removeLeaf (mx (i + 1) i)
                                <||> when (i == 2) (removeLeaf (polys i)))
               
----------------------------------------------------------------------
--- misc
----------------------------------------------------------------------                

checkArity = problems >>= mapM_ check
  where check p | null invalid = putStrLn "Arities are OK"
                | otherwise    = 
                  do putStrLn "Following symbols occur with miss-matching arities"
                     mapM_ (\ f -> pprint (f,sig)) invalid
        
          where sig = TR.signature p
                invalid = [ f | (f,i) <- symsWithArities
                              , (g,j) <- symsWithArities
                              , f == g , i /= j]
                symsWithArities = nub $ concatMap syms $ Set.toList $ TR.subterms $ TR.allComponents p
                nub = Set.toList . Set.fromList
                syms (TR.Var _) =  []
                syms (TR.Fun f ts) = [(f, length ts)]

